% learning Bayesian networks from data
% 1. creates the original network
% 2. samples the original network
% 3. takes samples and learns a new network from them
% a. the original network structure is known, only parameters are learnt
% b. the original network completely forgotten, only samples available, K2, MCMC and naive methods are demonstrated

% initialize FAMILY network (see the paper of Charniak and the lecture slides)
family_init;

% sample data from the original network
% forward sampling is used
fprintf(1,'Sampling the network ...\n');
nsamples=10000;
samples = cell(N, nsamples);
for i=1:nsamples
  samples(:,i) = sample_bnet(bnet_orig);
end

% 50% of sample values are removed -> incomplete data
samples2 = samples;
hide = rand(N, nsamples) > 0.5;
[I,J]=find(hide);
for k=1:length(I)
  samples2{I(k), J(k)} = [];
end

% parameter learning
% creates a new network, stems from the original FAMILY DAG,
% randomizes CPTs -- random values will later be changed for the learnt ones, 
% some CPT init is needed anyway (for more variables it shall be done in a for cycle)
bnet_rand = mk_bnet(dag, node_sizes,'names',names);
seed = 10;
rand('state', seed);
bnet_rand.CPD{FO} = tabular_CPD(bnet_rand, FO);
bnet_rand.CPD{LO} = tabular_CPD(bnet_rand, LO);
bnet_rand.CPD{BP} = tabular_CPD(bnet_rand, BP);
bnet_rand.CPD{DO} = tabular_CPD(bnet_rand, DO);
bnet_rand.CPD{HB} = tabular_CPD(bnet_rand, HB);

% apply ML to estimate parameter values  -- only a sample subset is used for rapid illustration
% print results for fo and do|fo,bp
fprintf(1,'Learning parameters from complete obs ...\n');
CPT_show = cell(1,N);
bnet_learn = learn_params(bnet_rand, samples(:,1:1000));
for j=1:N
    s=struct(bnet_learn.CPD{j});
    CPT_show{j}=s.CPT;
end
CPT_show{FO}
CPT_show{DO}

% a more readable CPT visualization
% family_init defines CPTs as follows: T~1 and F~2 (the common notatin F~0 and T~1 cannot be applied in BNT)
% the order of varaibles is: (FO BP: DO=T and DO=F)
% rodice jsou usporadani dle indexu, hodnoty uzlu jsou v poradi 1, 2
fprintf(1,'FO BP : DO=T DO=F\n');
dispcpt(CPT_show{DO});

% structure learning

% learn a network structure by K2 algorithm
fprintf(1,'Learning structure from complete obs by K2 ...\n');
order_user = [HB DO BP FO LO];
% order_simple = 1:N; % an arbitrary ordering can be tried
dagK2 = learn_struct_K2(samples, node_sizes, order_user, 'verbose','yes');
figure('Name','The best K2 structure');
draw_graph(dagK2,names);

% change the original variable order into a topological sort according to dagK2
% necessary e.g. for future application of jtree_inf_engine on this network (see BNT/examples/static/nodeorderExample.m)
% top_sort_dag is an external function, not a toolbox function
[dagK2_top order_K2 node_sizes_top names_K2] = top_sort_dag(dagK2,'node_sizes',node_sizes,'names',names);

bnet_K2 = mk_bnet(dagK2_top, node_sizes_top,'names',names_K2);
for i=1:N
    bnet_K2.CPD{i} = tabular_CPD(bnet_K2, i);
end

% learn network parameters -- the variable order must be changed according to a new sort
bnet_K2 = learn_params(bnet_K2, samples(order_K2,:));

% use the trained network for inference
engine = jtree_inf_engine(bnet_K2);
ev = cell(1,5);
ev{find(order_K2==HB)} = 2; % not heard bark, i.e. HB = F
ev{find(order_K2==LO)} = 1; % light is on, i.e. LO = T
engine = enter_evidence(engine, ev);
% when having the correct variable ordering (and the same ordering of variables in the sample):
% Pr(FO=T|HB=F,LO=T) approaches the original/real value of 0.5 (the estimate is a function of random sample and thus a random variable)
% without the topological sort, jtree returns remarkably different and incorrect estimates of Pr(FO=T|HB=F,LO=T)
m = marginal_nodes(engine, find(order_K2==FO));
fprintf('Pr(FO=T|HB=F,LO=T) = %5.4f\n', m.T(1))
fprintf('Pr(FO=F|HB=F,LO=T) = %5.4f\n', m.T(2))

% learn a network structure by Markov Chain Monte Carlo method
% demo for 100 samples only -- can be time consuming, specify internal params of MCMC as well to optimize
% (init_dag, nsamples etc.)
[sampled_graphs, accept_ratio] = learn_struct_mcmc(samples(:,1:100), node_sizes);
% network/graph evaluation
score = score_dags(samples(:,1:100), node_sizes, sampled_graphs);
% graph histogram according to evaluation function
figure('Name','Graph score histogram - MCMC without a starting graph');
bar(histc(score,unique(score)));
% show the graph with the best score
best_dag_indices = find(score == max(score));
best_dag_mcmc = sampled_graphs{1,best_dag_indices(1)};
figure('Name','The best MCMC graph');
draw_graph(best_dag_mcmc,names);
% MCMC convergency
% accept_ratio -- ratio of accepted network transitions (wrt proposed ones) in the window of last (N) networks, 
% in here it is expresed as the ratio of accepted to rejected transitions, the figure can be larger than 1 then, 
% the ideal accept_ratio is distribution dependent, e.g. for one-dimensional Gaussian distribution it is approximately 50%, it decreses to 23% for N-dimensional Gaussian 
fig_ar = figure('Name','Accept ratio');
plot(accept_ratio);

% exhaustive search of the graph space (naive solution)
% for 5 variables there is 29281 graphs ... not feasible
% let us omit HB, the number of graphs decreases to 543 then
all_4_dags=mk_all_dags(N-1);
% Bayesian criterion (default option in score_dags, set 'scoring_fn' to change it)
score=score_dags(samples(1:4,1:1000),node_sizes,all_4_dags);
best_dag_indices = find(score == max(score));
best_dag_exhaustive = all_4_dags{1,best_dag_indices(1)};
figure('Name','Exhaustive search, the best graf');
draw_graph(best_dag_exhaustive,names(1:4));
